@@ -10,9 +10,11 @@ import textwrap
 from distutils import log, ccompiler, sysconfig
 from distutils.core import Extension
 from distutils.ccompiler import get_default_compiler
+from importlib import invalidate_caches
 from setuptools.command.build_ext import build_ext as SetuptoolsBuildExt
+from setuptools.command import build_py
 
-from ..utils import get_numpy_include_path, invalidate_caches, classproperty
+from ..utils import get_numpy_include_path, classproperty
 from ..version_helpers import get_pkg_version_module
 
 
@@ -236,8 +238,7 @@ def generate_build_ext_command(packagename, release):
             # instance of a build_ext command that uses that base class (right
             # now the options being Cython.Distutils.build_ext, or the stock
             # setuptools build_ext)
-            new_cls = super(build_ext, cls._final_class).__new__(
-                    cls._final_class)
+            new_cls = super(build_ext, cls._final_class).__new__(cls._final_class)
 
             # Since the new cls is not a subclass of the original cls, we must
             # manually call its __init__
@@ -255,12 +256,14 @@ def generate_build_ext_command(packagename, release):
 
             extensions = self.distribution.ext_modules
             if extensions:
+                build_py = self.get_finalized_command('build_py')
+                package_dir = build_py.get_package_dir(packagename)
                 src_path = os.path.relpath(
                     os.path.join(os.path.dirname(__file__), 'src'))
                 shutil.copy(os.path.join(src_path, 'compiler.c'),
-                            os.path.join(self.package_name, '_compiler.c'))
+                            os.path.join(package_dir, '_compiler.c'))
                 ext = Extension(self.package_name + '._compiler',
-                                [os.path.join(self.package_name, '_compiler.c')])
+                                [os.path.join(package_dir, '_compiler.c')])
                 extensions.insert(0, ext)
 
             super(build_ext, self).finalize_options()
@@ -287,9 +290,11 @@ def generate_build_ext_command(packagename, release):
         def run(self):
             # For extensions that require 'numpy' in their include dirs,
             # replace 'numpy' with the actual paths
-            np_include = get_numpy_include_path()
+            np_include = None
             for extension in self.extensions:
                 if 'numpy' in extension.include_dirs:
+                    if np_include is None:
+                        np_include = get_numpy_include_path()
                     idx = extension.include_dirs.index('numpy')
                     extension.include_dirs.insert(idx, np_include)
                     extension.include_dirs.remove('numpy')
@@ -301,11 +306,12 @@ def generate_build_ext_command(packagename, release):
             # Update cython_version.py if building with Cython
             try:
                 cython_version = get_pkg_version_module(
-                        packagename, fromlist=['cython_version'])[0]
+                    packagename, fromlist=['cython_version'])[0]
             except (AttributeError, ImportError):
                 cython_version = 'unknown'
             if self.uses_cython and self.uses_cython != cython_version:
-                package_dir = os.path.relpath(packagename)
+                build_py = self.get_finalized_command('build_py')
+                package_dir = build_py.get_package_dir(packagename)
                 cython_py = os.path.join(package_dir, 'cython_version.py')
                 with open(cython_py, 'w') as f:
                     f.write('# Generated file; do not modify\n')
