@@ -4,19 +4,22 @@
 This is an Astropy affiliated package.
 """
 
-from __future__ import absolute_import, division, unicode_literals, print_function
+import sys
+if sys.version_info < (3, 5):
+    raise ImportError("GWCS does not support Python 2.x, 3.0, 3.1, 3.2, 3.3 or 3.4."
+                      "Beginning with GWCS 0.9, Python 3.5 and above is required.")
 
 
 # Affiliated packages may add whatever they like to this file, but
 # should keep this content at the top.
 # ----------------------------------------------------------------------------
-from ._astropy_init import *
+from ._astropy_init import * # noqa
 # ----------------------------------------------------------------------------
 
 
 # For egg_info test builds to pass, put package imports here.
 if not _ASTROPY_SETUP_:
-    from .wcs import *
-    from .wcstools import *
-    from .coordinate_frames import *
-    from .selector import *
+    from .wcs import *   # noqa
+    from .wcstools import *   # noqa
+    from .coordinate_frames import *  # noqa
+    from .selector import *   # noqa
