@@ -65,19 +65,21 @@ label mappers.
 
 
 """
-from __future__ import absolute_import, division, unicode_literals, print_function
-
+import warnings
 import numpy as np
 from astropy.modeling.core import Model
+from astropy.modeling import models as astmodels
 
 from . import region
 from .utils import RegionError, _toindex
 
 
-__all__ = ['LabelMapperArray', 'LabelMapperDict', 'LabelMapperRange', 'RegionsSelector']
+__all__ = ['LabelMapperArray', 'LabelMapperDict', 'LabelMapperRange', 'RegionsSelector',
+           'LabelMapper']
 
 
 def get_unique_regions(regions):
+    regions = np.asarray(regions)
     if isinstance(regions, np.ndarray):
         unique_regions = np.unique(regions).tolist()
 
@@ -184,11 +186,10 @@ class LabelMapperArray(_LabelMapper):
             _no_label = 0
         else:
             _no_label = ""
-
         super(LabelMapperArray, self).__init__(mapper, _no_label, name=name, **kwargs)
 
     def evaluate(self, *args):
-        args = [_toindex(a) for a in args]
+        args = tuple([_toindex(a) for a in args])
         try:
             result = self._mapper[args[::-1]]
         except IndexError as e:
@@ -224,11 +225,11 @@ class LabelMapperArray(_LabelMapper):
         Examples
         --------
         >>> regions = {1: [[795, 970], [2047, 970], [2047, 999], [795, 999], [795, 970]],
-                       2: [[844, 1067], [2047, 1067], [2047, 1113], [844, 1113], [844, 1067]],
-                       3: [[654, 1029], [2047, 1029], [2047, 1078], [654, 1078], [654, 1029]],
-                       4: [[772, 990], [2047, 990], [2047, 1042], [772, 1042], [772, 990]]
-                      }
-        >>> mapper = selector.LabelMapperArray.from_vertices((2400, 2400), regions)
+        ...            2: [[844, 1067], [2047, 1067], [2047, 1113], [844, 1113], [844, 1067]],
+        ...            3: [[654, 1029], [2047, 1029], [2047, 1078], [654, 1078], [654, 1029]],
+        ...            4: [[772, 990], [2047, 990], [2047, 1042], [772, 1042], [772, 990]]
+        ...           }
+        >>> mapper = LabelMapperArray.from_vertices((2400, 2400), regions)
 
         """
         labels = np.array(list(regions.keys()))
@@ -274,11 +275,21 @@ class LabelMapperDict(_LabelMapper):
 
     def __init__(self, inputs, mapper, inputs_mapping=None, atol=10**-8, name=None, **kwargs):
         self._atol = atol
-        self._inputs = inputs
         _no_label = 0
+        self._inputs = inputs
         if not all([m.n_outputs == 1 for m in mapper.values()]):
             raise TypeError("All transforms in mapper must have one output.")
-        super(LabelMapperDict, self).__init__(mapper, _no_label, inputs_mapping, name=name, **kwargs)
+        self._input_units_strict = {key: False for key in self._inputs}
+        self._input_units_allow_dimensionless = {key: False for key in self._inputs}
+        super(LabelMapperDict, self).__init__(mapper, _no_label, inputs_mapping,
+                                              name=name, **kwargs)
+
+    @property
+    def inputs(self):
+        """
+        The name(s) of the input variable(s) on which a model is evaluated.
+        """
+        return self._inputs
 
     @property
     def atol(self):
@@ -288,10 +299,6 @@ class LabelMapperDict(_LabelMapper):
     def atol(self, val):
         self._atol = val
 
-    @property
-    def inputs(self):
-        return self._inputs
-
     def evaluate(self, *args):
         shape = args[0].shape
         args = [a.flatten() for a in args]
@@ -302,11 +309,10 @@ class LabelMapperDict(_LabelMapper):
             keys = args
         keys = keys.flatten()
         # create an empty array for the results
-        res = np.empty(keys.shape)
+        res = np.zeros(keys.shape) + self._no_label
         # If this is part of a combined transform, some of the inputs
         # may be NaNs.
         # Set NaNs to the ``_no_label`` value
-        res[np.isnan(keys)] = self._no_label
         mapper_keys = list(self.mapper.keys())
         # Loop over the keys in mapper and compare to inputs.
         # Find the indices where they are within ``atol``
@@ -358,7 +364,17 @@ class LabelMapperRange(_LabelMapper):
         _no_label = 0
         if not all([m.n_outputs == 1 for m in mapper.values()]):
             raise TypeError("All transforms in mapper must have one output.")
-        super(LabelMapperRange, self).__init__(mapper, _no_label, inputs_mapping, name=name, **kwargs)
+        self._input_units_strict = {key: False for key in self._inputs}
+        self._input_units_allow_dimensionless = {key: False for key in self._inputs}
+        super(LabelMapperRange, self).__init__(mapper, _no_label, inputs_mapping,
+                                               name=name, **kwargs)
+
+    @property
+    def inputs(self):
+        """
+        The name(s) of the input variable(s) on which a model is evaluated.
+        """
+        return self._inputs
 
     @staticmethod
     def _has_overlapping(ranges):
@@ -375,7 +391,7 @@ class LabelMapperRange(_LabelMapper):
         l = np.array(l)
         start = np.roll(l[:, 0], -1)
         end = l[:, 1]
-        if any((end - start)[:-1]  > 0) or any(start[-1] > end):
+        if any((end - start)[:-1] > 0) or any(start[-1] > end):
             return True
         else:
             return False
@@ -407,10 +423,6 @@ class LabelMapperRange(_LabelMapper):
         else:
             return ind.item()
 
-    @property
-    def inputs(self):
-        return self._inputs
-
     def evaluate(self, *args):
         shape = args[0].shape
         args = [a.flatten() for a in args]
@@ -420,7 +432,7 @@ class LabelMapperRange(_LabelMapper):
             keys = args
         keys = keys.flatten()
         # Define an array for the results.
-        res = np.empty(keys.shape)
+        res = np.zeros(keys.shape) + self._no_label
         nan_ind = np.isnan(keys)
         res[nan_ind] = self._no_label
         value_ranges = list(self.mapper.keys())
@@ -428,17 +440,20 @@ class LabelMapperRange(_LabelMapper):
         # which fall within the range it defines.
         for val_range in value_ranges:
             temp = keys.copy()
-            temp[nan_ind] = self._no_label
+            temp[nan_ind] = np.nan
             temp = np.where(np.logical_or(temp <= val_range[0],
                                           temp >= val_range[1]),
-                                          0, temp)
-            ind = temp.nonzero()
-            if ind:
+                            np.nan, temp)
+            ind = ~np.isnan(temp)
+
+            if ind.any():
                 inputs = [a[ind] for a in args]
                 res[ind] = self.mapper[tuple(val_range)](*inputs)
             else:
                 continue
         res.shape = shape
+        if len(np.nonzero(res)[0]) == 0:
+            warnings.warn("All data is outside the valid range - {0}.".format(self.name))
         return res
 
 
@@ -482,7 +497,8 @@ class RegionsSelector(Model):
 
         if " " in selector.keys() or 0 in selector.keys():
             raise ValueError('"0" and " " are not allowed as keys.')
-
+        self._input_units_strict = {key: False for key in self._inputs}
+        self._input_units_allow_dimensionless = {key: False for key in self._inputs}
         super(RegionsSelector, self).__init__(n_models=1, name=name, **kwargs)
 
     def set_input(self, rid):
@@ -521,7 +537,7 @@ class RegionsSelector(Model):
         rids = self.label_mapper(*args).flatten()
         # Raise an error if all pixels are outside regions
         if (rids == self.label_mapper.no_label).all():
-            raise RegionError("The input positions are not inside any region.")
+            warnings.warn("The input positions are not inside any region.")
 
         # Create output arrays and set any pixels not within regions to
         # "undefined_transform_value"
@@ -537,7 +553,12 @@ class RegionsSelector(Model):
         for rid in uniq:
             ind = (rids == rid)
             inputs = [a[ind] for a in args]
-            result = self._selector[rid](*inputs)
+            if rid in self._selector:
+                result = self._selector[rid](*inputs)
+            else:
+                # If there's no transform for a label, return np.nan
+                result = [np.empty(inputs[0].shape) +
+                          self._undefined_transform_value for i in range(self.n_outputs)]
             for j in range(self.n_outputs):
                 outputs[j][ind] = result[j]
         return outputs
@@ -550,13 +571,6 @@ class RegionsSelector(Model):
     def undefined_transform_value(self, value):
         self._undefined_transform_value = value
 
-    @property
-    def inputs(self):
-        """
-        The name(s) of the input variable(s) on which a model is evaluated.
-        """
-        return self._inputs
-
     @property
     def outputs(self):
         """The name(s) of the output(s) of the model."""
@@ -565,3 +579,79 @@ class RegionsSelector(Model):
     @property
     def selector(self):
         return self._selector
+
+    @property
+    def inputs(self):
+        """
+        The name(s) of the input variable(s) on which a model is evaluated.
+        """
+        return self._inputs
+
+
+class LabelMapper(_LabelMapper):
+    """
+    Maps inputs to regions. Returns the region labels corresponding to the inputs.
+
+    Labels are strings or numbers which uniquely identify a location.
+    For example, labels may represent slices of an IFU or names of spherical polygons.
+
+    Parameters
+    ----------
+    mapper : `~astropy.modeling.core.Model`
+        A function which returns a region.
+    no_label : str or int
+        "" or 0
+        A return value for a location which has no corresponding label.
+    inputs_mapping : `~astropy.modeling.mappings.Mapping` or tuple
+        An optional Mapping model to be prepended to the LabelMapper
+        with the purpose to filter the inputs or change their order.
+        If tuple, a `~astropy.modeling.mappings.Mapping` model will be created from it.
+    name : str
+        The name of this transform.
+    """
+
+    outputs = ('label',)
+
+    def __init__(self, inputs, mapper, no_label=np.nan, inputs_mapping=None, name=None, **kwargs):
+        self._no_label = no_label
+        self._inputs = inputs
+        self.outputs = tuple(['x{0}'.format(ind) for ind in list(range(mapper.n_outputs))])
+        if isinstance(inputs_mapping, tuple):
+            inputs_mapping = astmodels.Mapping(inputs_mapping)
+        elif inputs_mapping is not None and not isinstance(inputs_mapping, astmodels.Mapping):
+            raise TypeError("inputs_mapping must be an instance of astropy.modeling.Mapping.")
+
+        self._inputs_mapping = inputs_mapping
+        self._mapper = mapper
+        self._input_units_strict = {key: False for key in self._inputs}
+        self._input_units_allow_dimensionless = {key: False for key in self._inputs}
+        super(_LabelMapper, self).__init__(name=name, **kwargs)
+
+    @property
+    def inputs(self):
+        """
+        The name(s) of the input variable(s) on which a model is evaluated.
+        """
+        return self._inputs
+
+    @property
+    def mapper(self):
+        return self._mapper
+
+    @property
+    def inputs_mapping(self):
+        return self._inputs_mapping
+
+    @property
+    def no_label(self):
+        return self._no_label
+
+    def evaluate(self, *args):
+        if self.inputs_mapping is not None:
+            args = self.inputs_mapping(*args)
+        if self.n_outputs == 1:
+            args = [args]
+        res = self.mapper(*args)
+        if np.isscalar(res):
+            res = np.array([res])
+        return np.array(res)
