@@ -1,6 +1,5 @@
 # Licensed under a 3-clause BSD style license - see LICENSE.rst
-"""
-ASDF readers/writers
-"""
-from .selectortags import *
+# -*- coding: utf-8 -*-
 
+from .wcs import *
+from .selectortags import *
