@@ -49,17 +49,6 @@ class CoordinateFrameError(Exception):
         super(CoordinateFrameError, self).__init__(message)
 
 
-def uses_quantity(transform):
-    """
-    True if this model has been created with `~astropy.units.Quantity`
-    objects or if there are no parameters.
-    This can be used to determine if this model should be evaluated with
-    `~astropy.units.Quantity` or regular floats.
-    """
-    pisq = [isinstance(p, u.Quantity) for p in transform._param_sets(units=True)]
-    return (len(pisq) == 0) or any(pisq)
-
-
 def _toindex(value):
     """
     Convert value to an int or an int array.
@@ -179,7 +168,7 @@ def read_wcs_from_header(header):
     try:
         wcs_info['WCSAXES'] = header['WCSAXES']
     except KeyError:
-        p = re.compile('ctype[\d]*', re.IGNORECASE)
+        p = re.compile(r'ctype[\d]*', re.IGNORECASE)
         ctypes = header['CTYPE*']
         keys = list(ctypes.keys())
         for key in keys[::-1]:
@@ -477,50 +466,3 @@ def isnumerical(val):
     elif isinstance(val, np.ndarray) and val.dtype not in dtypes:
         isnum = False
     return isnum
-
-
-def separable_axes(wcsobj, start_frame=None, end_frame=None):
-        """
-        Computes the separability of axes in ``end_frame``.
-
-        Returns a 1D boolean array of size frame.naxes where True means
-        the axis is completely separable and False means the axis is nonseparable
-        from at least one other axis.
-
-        Parameters
-        ----------
-        wcsobj : `~gwcs.wcs.WCS`
-            WCS object
-        start_frame : `~gwcs.coordinate_frames.CoordinateFrame`
-            A frame in the WCS pipeline.
-            The transform between start_frame and the end frame is used to compute the
-            mapping inputs: outputs.
-            If None the input_frame is used as start_frame.
-        end_frame : `~gwcs.coordinate_frames.CoordinateFrame`
-            A frame in the WCS pipeline.
-            The transform between start_frame and the end frame is used to compute the
-            mapping inputs: outputs.
-            If None wcsobj.output_frame is used.
-
-        See Also
-        --------
-        input_axes : For each output axis return the input axes contributing to it.
-
-        """
-        if wcsobj is not None:
-            if start_frame is None:
-                start_frame = wcsobj.input_frame
-            else:
-                if start_frame not in wcsobj.available_frames:
-                    raise ValueError("Unrecognized frame {0}".format(start_frame))
-            if end_frame is None:
-                end_frame = wcsobj.output_frame
-            else:
-                if end_frame not in wcsobj.available_frames:
-                    raise ValueError("Unrecognized frame {0}".format(end_frame))
-            transform = wcsobj.get_transform(start_frame, end_frame)
-        else:
-            raise ValueError("A starting frame is needed to determine separability of axes.")
-
-        sep = is_separable(transform)
-        return [sep[ax] for ax in end_frame.axes_order]
