@@ -1,21 +1,21 @@
 # Licensed under a 3-clause BSD style license - see LICENSE.rst
-from __future__ import absolute_import, division, unicode_literals, print_function
-
 import functools
+import itertools
 import numpy as np
-from astropy.extern import six
 from astropy.modeling.core import Model
+from astropy.modeling import utils as mutils
 
+from .api import GWCSAPIMixin
 from . import coordinate_frames
 from .utils import CoordinateFrameError
 from .utils import _toindex
 from . import utils
 
+
 __all__ = ['WCS']
 
 
-class WCS(object):
-
+class WCS(GWCSAPIMixin):
     """
     Basic WCS class.
 
@@ -37,9 +37,11 @@ class WCS(object):
 
     def __init__(self, forward_transform=None, input_frame='detector', output_frame=None,
                  name=""):
+        #self.low_level_wcs = self
         self._available_frames = []
         self._pipeline = []
         self._name = name
+        self._array_shape = None
         self._initialize_wcs(forward_transform, input_frame, output_frame)
 
     def _initialize_wcs(self, forward_transform, input_frame, output_frame):
@@ -54,14 +56,14 @@ class WCS(object):
                 super(WCS, self).__setattr__(_input_frame, inp_frame_obj)
                 super(WCS, self).__setattr__(_output_frame, outp_frame_obj)
 
-                self._pipeline = [(_input_frame, forward_transform.copy()),
-                                  (_output_frame, None)]
+                self._pipeline = [(input_frame, forward_transform.copy()),
+                                  (output_frame, None)]
             elif isinstance(forward_transform, list):
                 for item in forward_transform:
                     name, frame_obj = self._get_frame_name(item[0])
                     super(WCS, self).__setattr__(name, frame_obj)
-                    self._pipeline.append((name, item[1]))
-
+                    #self._pipeline.append((name, item[1]))
+                    self._pipeline = forward_transform
             else:
                 raise TypeError("Expected forward_transform to be a model or a "
                                 "(frame, transform) list, got {0}".format(
@@ -99,7 +101,8 @@ class WCS(object):
         try:
             from_ind = self._get_frame_index(from_frame)
         except ValueError:
-            raise CoordinateFrameError("Frame {0} is not in the available frames".format(from_frame))
+            raise CoordinateFrameError("Frame {0} is not in the available "
+                                       "frames".format(from_frame))
         try:
             to_ind = self._get_frame_index(to_frame)
         except ValueError:
@@ -183,8 +186,9 @@ class WCS(object):
         """
         if isinstance(frame, coordinate_frames.CoordinateFrame):
             frame = frame.name
-        return np.asarray(self._pipeline)[:, 0].tolist().index(frame)
-
+        frame_names = [getattr(item[0], "name", item[0]) for item in self._pipeline]
+        return frame_names.index(frame)
+        
     def _get_frame_name(self, frame):
         """
         Return the name of the frame and a ``CoordinateFrame`` object.
@@ -201,7 +205,7 @@ class WCS(object):
         frame_obj : `~gwcs.coordinate_frames.CoordinateFrame`
             Frame instance or None (if `frame` is str)
         """
-        if isinstance(frame, six.string_types):
+        if isinstance(frame, str):
             name = frame
             frame_obj = None
         else:
@@ -214,28 +218,53 @@ class WCS(object):
         Executes the forward transform.
 
         args : float or array-like
-            Inputs in the input coordinate system, separate inputs for each dimension.
-        output : str
-            One of ["numericals", "numericals_plus"]
-            If "numericals_plus" - returns a `~astropy.coordinates.SkyCoord` or
-            `~astropy.units.Quantity` object.
+            Inputs in the input coordinate system, separate inputs
+            for each dimension.
+        with_units : bool
+            If ``True`` returns a `~astropy.coordinates.SkyCoord` or
+            `~astropy.units.Quantity` object, by using the units of
+            the output cooridnate frame.
+            Optional, default=False.
+        with_bounding_box : bool, optional
+             If True(default) values in the result which correspond to
+             any of the inputs being outside the bounding_box are set
+             to ``fill_value``.
+        fill_value : float, optional
+            Output value for inputs outside the bounding_box
+            (default is np.nan).
         """
-        if self.forward_transform is None:
+        transform = self.forward_transform
+        if transform is None:
             raise NotImplementedError("WCS.forward_transform is not implemented.")
-        result = self.forward_transform(*args)
-        output = kwargs.pop('output', None)
-        if output == 'numericals_plus':
+
+        with_units = kwargs.pop("with_units", False)
+        if 'with_bounding_box' not in kwargs:
+            kwargs['with_bounding_box'] = True
+        if 'fill_value' not in kwargs:
+            kwargs['fill_value'] = np.nan
+
+        if self.bounding_box is not None:
+            # Currently compound models do not attempt to combine individual model
+            # bounding boxes. Get the forward transform and assign the ounding_box to it
+            # before evaluating it. The order Model.bounding_box is reversed.
+            axes_ind = self._get_axes_indices()
+            if transform.n_inputs > 1:
+                transform.bounding_box = np.array(self.bounding_box)[axes_ind][::-1]
+            else:
+                transform.bounding_box = self.bounding_box
+        result = transform(*args, **kwargs)
+
+        if with_units:
             if self.output_frame.naxes == 1:
                 result = self.output_frame.coordinates(result)
             else:
                 result = self.output_frame.coordinates(*result)
-        elif output is not None and output != "numericals":
-            raise ValueError("Type of output unrecognized {0}".format(output))
+
         return result
 
     def invert(self, *args, **kwargs):
         """
-        Invert coordnates.
+        Invert coordinates.
 
         The analytical inverse of the forward transform is used, if available.
         If not an iterative method is used.
@@ -246,15 +275,29 @@ class WCS(object):
             coordinates to be inverted
         kwargs : dict
             keyword arguments to be passed to the iterative invert method.
+        with_bounding_box : bool, optional
+             If True(default) values in the result which correspond to any of the inputs being
+             outside the bounding_box are set to ``fill_value``.
+        fill_value : float, optional
+            Output value for inputs outside the bounding_box (default is np.nan).
         """
         if not utils.isnumerical(args[0]):
-            args = utils._get_values(self.unit, *args)
+            args = self.output_frame.coordinate_to_quantity(*args)
+            if not self.forward_transform.uses_quantity:
+                args = utils.get_values(self.output_frame.unit, *args)
+
+        with_units = kwargs.pop('with_units', False)
+        if 'with_bounding_box' not in kwargs:
+            kwargs['with_bounding_box'] = True
+        if 'fill_value' not in kwargs:
+            kwargs['fill_value'] = np.nan
+
         try:
-            result = self.backward_transform(*args)
+            result = self.backward_transform(*args, **kwargs)
         except (NotImplementedError, KeyError):
             result = self._invert(*args, **kwargs)
-        output = kwargs.pop('output', None)
-        if output == 'numericals_plus':
+
+        if with_units and self.input_frame:
             if self.input_frame.naxes == 1:
                 return self.input_frame.coordinates(result)
             else:
@@ -278,22 +321,33 @@ class WCS(object):
             Initial coordinate frame.
         to_frame : str, or instance of `~gwcs.cordinate_frames.CoordinateFrame`
             Coordinate frame into which to transform.
-        args : float
-            input coordinates to transform
         args : float or array-like
             Inputs in ``from_frame``, separate inputs for each dimension.
-        output : str
-            One of ["numericals", "numericals_plus"]
-            If "numericals_plus" - returns a `~astropy.coordinates.SkyCoord` or
+        output_with_units : bool
+            If ``True`` - returns a `~astropy.coordinates.SkyCoord` or
             `~astropy.units.Quantity` object.
+        with_bounding_box : bool, optional
+             If True(default) values in the result which correspond to any of the inputs being
+             outside the bounding_box are set to ``fill_value``.
+        fill_value : float, optional
+            Output value for inputs outside the bounding_box (default is np.nan).
         """
         transform = self.get_transform(from_frame, to_frame)
         if not utils.isnumerical(args[0]):
-            args = utils._get_values(self.unit, *args)
+            inp_frame = getattr(self, from_frame)
+            args = inp_frame.coordinate_to_quantity(*args)
+            if not transform.uses_quantity:
+                args = utils.get_values(inp_frame.unit, *args)
 
-        result = transform(*args)
-        output = kwargs.pop("output", None)
-        if output == "numericals_plus":
+        with_units = kwargs.pop("with_units", False)
+        if 'with_bounding_box' not in kwargs:
+            kwargs['with_bounding_box'] = True
+        if 'fill_value' not in kwargs:
+            kwargs['fill_value'] = np.nan
+
+        result = transform(*args, **kwargs)
+
+        if with_units:
             to_frame_name, to_frame_obj = self._get_frame_name(to_frame)
             if to_frame_obj is not None:
                 if to_frame_obj.naxes == 1:
@@ -303,8 +357,7 @@ class WCS(object):
             else:
                 raise TypeError("Coordinate objects could not be created because"
                                 "frame {0} is not defined.".format(to_frame_name))
-        elif output is not None and output != "numericals":
-            raise ValueError("Type of output unrecognized {0}".format(output))
+
         return result
 
     @property
@@ -318,7 +371,7 @@ class WCS(object):
             {frame_name: frame_object or None}
         """
         if self._pipeline:
-            return [frame[0] for frame in self._pipeline]
+            return [getattr(frame[0], "name", frame[0]) for frame in self._pipeline]
         else:
             return None
 
@@ -336,7 +389,7 @@ class WCS(object):
             New transform to be inserted in the pipeline
         after : bool
             If True, the new transform is inserted in the pipeline
-            immediately after `frame`.
+            immediately after ``frame``.
         """
         name, _ = self._get_frame_name(frame)
         frame_ind = self._get_frame_index(name)
@@ -352,7 +405,7 @@ class WCS(object):
         """The unit of the coordinates in the output coordinate system."""
         if self._pipeline:
             try:
-                return getattr(self, self._pipeline[-1][0]).unit
+                return getattr(self, self._pipeline[-1][0].name).unit
             except AttributeError:
                 return None
         else:
@@ -362,7 +415,10 @@ class WCS(object):
     def output_frame(self):
         """Return the output coordinate frame."""
         if self._pipeline:
-            return getattr(self, self._pipeline[-1][0])
+            frame = self._pipeline[-1][0]
+            if not isinstance(frame, str):
+                frame = frame.name
+            return getattr(self, frame)
         else:
             return None
 
@@ -370,7 +426,10 @@ class WCS(object):
     def input_frame(self):
         """Return the input coordinate frame."""
         if self._pipeline:
-            return getattr(self, self._pipeline[0][0])
+            frame = self._pipeline[0][0]
+            if not isinstance(frame, str):
+                frame = frame.name
+            return getattr(self, frame)
         else:
             return None
 
@@ -390,40 +449,66 @@ class WCS(object):
         return self._pipeline
 
     @property
-    def domain(self):
+    def bounding_box(self):
         """
         Return the range of acceptable values for each input axis.
+        The order of the axes is `~gwcs.coordinate_frames.CoordinateFrame.axes_order`.
         """
         frames = self.available_frames
-        transform_meta = self.get_transform(frames[0], frames[1]).meta
-        if 'domain' in transform_meta:
-            return transform_meta['domain']
-        else:
+        transform_0 = self.get_transform(frames[0], frames[1])
+        try:
+            bb = transform_0.bounding_box
+        except NotImplementedError:
             return None
+        if transform_0.n_inputs == 1:
+            return bb
+        try:
+            axes_order = self.input_frame.axes_order
+        except AttributeError:
+            axes_order = np.arange(transform_0.n_inputs)
+        # Model.bounding_box is in python order, need to reverse it first.
+        bb = np.array(bb[::-1])[np.array(axes_order)]
+        return tuple(tuple(item) for item in bb)
 
-    @domain.setter
-    def domain(self, value):
+    @bounding_box.setter
+    def bounding_box(self, value):
         """
         Set the range of acceptable values for each input axis.
-        """
-        self._validate_domain(value)
-        frames = self.available_frames
-        transform = self.get_transform(frames[0], frames[1])
-        transform.meta['domain'] = value
-        self.set_transform(frames[0], frames[1], transform)
 
-    def _validate_domain(self, domain):
-        n_inputs = self.forward_transform.n_inputs
-        if len(domain) != n_inputs:
-            raise ValueError("The number of domains should match the number "
-                             "of inputs {0}".format(n_inputs))
-        if not isinstance(domain, (list, tuple)) or \
-           not all([isinstance(d, dict) for d in domain]):
-            raise TypeError('"domain" should be a list of dictionaries for each axis in the input_frame'
-                            "[{'lower': low_x, "
-                            "'upper': high_x, "
-                            "'includes_lower': bool, "
-                            "'includes_upper': bool}]")
+        The order of the axes is `~gwcs.coordinate_frames.CoordinateFrame.axes_order`.
+        For two inputs and axes_order(0, 1) the bounding box is ((xlow, xhigh), (ylow, yhigh)).
+
+        Parameters
+        ----------
+        value : tuple or None
+            Tuple of tuples with ("low", high") values for the range.
+        """
+        frames = self.available_frames
+        transform_0 = self.get_transform(frames[0], frames[1])
+        if value is None:
+            transform_0.bounding_box = value
+        else:
+            try:
+                # Make sure the dimensions of the new bbox are correct.
+                mutils._BoundingBox.validate(transform_0, value)
+            except:
+                raise
+            # get the sorted order of axes' indices
+            axes_ind = self._get_axes_indices()
+            if transform_0.n_inputs == 1:
+                transform_0.bounding_box = value
+            else:
+                # The axes in bounding_box in modeling follow python order
+                transform_0.bounding_box = np.array(value)[axes_ind][::-1]
+        self.set_transform(frames[0], frames[1], transform_0)
+
+    def _get_axes_indices(self):
+        try:
+            axes_ind = np.argsort(self.input_frame.axes_order)
+        except AttributeError:
+            # the case of a frame being a string
+            axes_ind = np.arange(self.forward_transform.n_inputs)
+        return axes_ind
 
     def __str__(self):
         from astropy.table import Table
@@ -444,33 +529,65 @@ class WCS(object):
             self.output_frame, self.input_frame, self.forward_transform)
         return fmt
 
-    def footprint(self, domain=None, center=True):
+    def footprint(self, bounding_box=None, center=False, axis_type="all"):
         """
-        Return the footprint of the observation in world coordinates.
+        Return the footprint in world coordinates.
 
         Parameters
         ----------
-        domain : slice or tuple of floats: (start, stop, step) or (start, stop) or (stop,)
-            size of image
+        bounding_box : tuple of floats: (start, stop)
+            `prop: bounding_box`
         center : bool
             If `True` use the center of the pixel, otherwise use the corner.
+        axis_type : str
+            A supported ``output_frame.axes_type`` or "all" (default).
+            One of ['spatial', 'spectral', 'temporal'] or a custom type.
 
         Returns
         -------
-        coord : (4, 2) array of (*x*, *y*) coordinates.
-            The order is counter-clockwise starting with the bottom left corner.
-        """
-        if domain is None:
-            if self.domain is None:
-                raise TypeError("Need a valid domain to compute the footprint.")
-            domain = self.domain
-        self._validate_domain(domain)
+        coord : ndarray
+            Array of coordinates in the output_frame mapping
+            corners to the output frame. For spatial coordinates the order
+            is clockwise, starting from the bottom left corner.
 
-        bounds = utils._domain_to_bounds(domain)
-        vertices = np.asarray([[bounds[0][0], bounds[1][0]], [bounds[0][0], bounds[1][1]],
-                               [bounds[0][1], bounds[1][1]], [bounds[0][1], bounds[1][0]]])
-        vertices = _toindex(vertices).T
-        if not center:
-            vertices += .5
-        result = self.__call__(*vertices)
-        return np.asarray(result)
+        """
+        def _order_clockwise(v):
+            return np.asarray([[v[0][0], v[1][0]], [v[0][0], v[1][1]],
+                               [v[0][1], v[1][1]], [v[0][1], v[1][0]]]).T
+
+        if bounding_box is None:
+            if self.bounding_box is None:
+                raise TypeError("Need a valid bounding_box to compute the footprint.")
+            bb = self.bounding_box
+        else:
+            bb = bounding_box
+
+        all_spatial = all([t.lower() == "spatial" for t in self.output_frame.axes_type])
+
+        if all_spatial:
+            vertices = _order_clockwise(bb)
+        else:
+            vertices = np.array(list(itertools.product(*bb))).T
+
+        if center:
+            vertices = _toindex(vertices)
+
+        result = np.asarray(self.__call__(*vertices, **{'with_bounding_box': False}))
+
+        axis_type = axis_type.lower()
+        if axis_type == 'spatial' and all_spatial:
+            return result.T
+
+        if axis_type != "all":
+            axtyp_ind = np.array([t.lower() for t in self.output_frame.axes_type]) == axis_type
+            if not axtyp_ind.any():
+                raise ValueError('This WCS does not have axis of type "{}".'.format(axis_type))
+            result = np.asarray([(r.min(), r.max()) for r in result[axtyp_ind]])
+
+            if axis_type == "spatial":
+                result = _order_clockwise(result)
+            else:
+                result.sort()
+                result = np.squeeze(result)
+
+        return result.T
